@@ -3,7 +3,8 @@
  *
  * This file is part of msmtp, an SMTP client, and of mpop, a POP3 client.
  *
- * Copyright (C) 2004, 2005, 2006, 2007, 2009, 2011, 2014, 2018, 2019, 2020
+ * Copyright (C) 2004, 2005, 2006, 2007, 2009, 2011, 2014, 2018, 2019, 2020,
+ * 2021, 2022, 2023
  * Martin Lambers <marlam@marlam.de>
  *
  *   This program is free software; you can redistribute it and/or modify
@@ -47,8 +48,13 @@
 #include <sys/stat.h>
 #include <time.h>
 #include <unistd.h>
+#ifdef ENABLE_NLS
+# include <locale.h>
+# include <langinfo.h>
+#endif
 
 #include "xalloc.h"
+#include "base64.h"
 #include "tools.h"
 
 
@@ -763,6 +769,37 @@ char *sanitize_string(char *str)
     return str;
 }
 
+/*
+ * token_in_string()
+ *
+ * see tools.h
+ */
+
+int token_in_string(const char *string, const char *token)
+{
+    size_t token_len = strlen(token);
+    const char *oldstring = string;
+    const char *newstring;
+    int found = 0;
+
+    while ((newstring = strstr(oldstring, token)))
+    {
+        size_t i = newstring - string;
+        if (i == 0 || string[i - 1] == ' ') /* valid start of token */
+        {
+            if (string[i + token_len] == ' '
+                    || string[i + token_len] == '\r'
+                    || string[i + token_len] == '\0') /* valid end of token */
+            {
+                found = 1;
+                break;
+            }
+        }
+        oldstring = newstring + 1;
+    }
+    return found;
+}
+
 
 /*
  * print_fingerprint()
@@ -792,49 +829,20 @@ void print_fingerprint(char *s, const unsigned char *fingerprint, size_t len)
 
 void print_time_rfc2822(time_t t, char rfc2822_timestamp[32])
 {
-    /* Calculate a RFC 2822 timestamp. strftime() is unreliable for this because
-     * it is locale dependant, and because the timezone offset conversion
-     * specifier %z is not portable. */
-    struct tm gmt, *lt;
-    char tz_offset_sign;
-    int tz_offset_hours;
-    int tz_offset_minutes;
-    const char *weekday[7] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri",
-        "Sat" };
-    const char *month[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
-        "Aug", "Sep", "Oct", "Nov", "Dec" };
-
-    /* copy the struct tm, because the subsequent call to localtime() will
-     * overwrite it */
-    gmt = *gmtime(&t);
-    lt = localtime(&t);
-    tz_offset_minutes = (lt->tm_hour - gmt.tm_hour) * 60
-        + lt->tm_min - gmt.tm_min
-        + (lt->tm_year - gmt.tm_year) * 24 * 60
-        + (lt->tm_yday - gmt.tm_yday) * 24 * 60;
-    if (tz_offset_minutes < 0)
-    {
-        tz_offset_sign = '-';
-        tz_offset_minutes = -tz_offset_minutes;
-    }
-    else
-    {
-        tz_offset_sign = '+';
-    }
-    tz_offset_hours = tz_offset_minutes / 60;
-    tz_offset_minutes %= 60;
-    if (tz_offset_hours > 99)
-    {
-        /* Values equal to or larger than 24 are not meaningful, but we just
-         * make sure that the value fits into two digits. If the system time is
-         * broken, we cannot fix it. */
-        tz_offset_hours = 99;
-    }
-    (void)snprintf(rfc2822_timestamp, 32,
-            "%s, %02d %s %04d %02d:%02d:%02d %c%02d%02d",
-            weekday[lt->tm_wday], lt->tm_mday, month[lt->tm_mon],
-            lt->tm_year + 1900, lt->tm_hour, lt->tm_min, lt->tm_sec,
-            tz_offset_sign, tz_offset_hours, tz_offset_minutes);
+    struct tm *lt = localtime(&t);
+#ifdef ENABLE_NLS
+    /* Set the correct locale for strftime() */
+    char *old_locale, *saved_locale;
+    old_locale = setlocale(LC_ALL, NULL);
+    saved_locale = xstrdup(old_locale);
+    setlocale(LC_ALL, "C");
+#endif
+    strftime(rfc2822_timestamp, 32, "%a, %d %b %Y %T %z", lt);
+#ifdef ENABLE_NLS
+    /* Restore the original locale */
+    setlocale(LC_ALL, saved_locale);
+    free(saved_locale);
+#endif
 }
 
 
@@ -901,3 +909,55 @@ int check_hostname_matches_domain(const char *hostname, const char *domain)
     return (hostname[hostname_len - 1 - domain_len] == '.'
             && strcasecmp(hostname + (hostname_len - domain_len), domain) == 0) ? 1 : 0;
 }
+
+
+/*
+ * encode_for_header()
+ *
+ * see tools.h
+ */
+
+char *encode_for_header(const char *s)
+{
+    int needsEncoding = 0;
+    for (int i = 0; s[i]; i++)
+    {
+        if (s[i] < 32 || s[i] >= 127)
+        {
+            needsEncoding = 1;
+            break;
+        }
+    }
+    if (needsEncoding)
+    {
+        /* create a string of the form "=?ENCODING?B?BASE64STRING?=" */
+        size_t s_len = strlen(s);
+        size_t b64_s_len = BASE64_LENGTH(s_len);
+        char* encoding =
+#ifdef ENABLE_NLS
+                nl_langinfo(CODESET);
+#else
+                "UTF-8";
+#endif
+        size_t e_len = strlen(encoding);
+        size_t enc_len = 2 + e_len + 3 + b64_s_len + 3;
+        char *enc = xmalloc(enc_len + 1);
+        size_t i = 0;
+        enc[i++] = '=';
+        enc[i++] = '?';
+        for (size_t j = 0; j < e_len; j++)
+        {
+            enc[i++] = tolower(encoding[j]);
+        }
+        enc[i++] = '?';
+        enc[i++] = 'B';
+        enc[i++] = '?';
+        base64_encode(s, s_len, enc + i, enc_len - i + 1);
+        strcat(enc, "?=");
+        return enc;
+    }
+    else
+    {
+        return xstrdup(s);
+    }
+}
